/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>


#include "initng_iowait.h"

void iowait_init(void)
{
    null_timezone.tz_minuteswest = 0;
    null_timezone.tz_dsttime = 0;
    proc_stat_fd = 0;

    gettimeofday(&current_time, &null_timezone);
    old_time.tv_sec = current_time.tv_sec;
    old_time.tv_usec = current_time.tv_usec;
    current_time.tv_sec = 0;                /* set this to make first check pass */
    CLEARNUMBERS;                           /* dont collectstat here, it WILL fail */
}

int do_recheck(void)
{
    gettimeofday(&current_time, &null_timezone);
    if (old_time.tv_sec == current_time.tv_sec
        && (current_time.tv_usec - old_time.tv_usec < MIN_PROCTIME))
    {
        return FALSE;
    }
    old_time.tv_sec = current_time.tv_sec;
    old_time.tv_usec = current_time.tv_usec;
    return TRUE;
}

void collectstat(void)
{
    char buf[127];

    unsigned long long x;       /* used with set_procstat_value define */

    if (proc_stat_fd < 1)
        proc_stat_fd = open(PROC_STAT, O_RDONLY);       /* todo, keep fd open and lseek it */
    else
        lseek(proc_stat_fd, 0L, SEEK_SET);
    if (proc_stat_fd < 1)
    {
        CLEARNUMBERS;
        printf("Failed to open %s\n", PROC_STAT);
        return;
    }
    if (read(proc_stat_fd, buf, 127) < 0)
    {
        CLEARNUMBERS;
        close(proc_stat_fd);
        proc_stat_fd = 0;
        printf("Failed to read from %s\n", PROC_STAT);
        return;
    }
    sscanf(buf, "cpu  %Lu %Lu %Lu %Lu %Lu %Lu %Lu", &current_cpuuser,
           &current_cpunice, &current_cpusystem, &current_cpuidle,
           &current_cpuiowait, &current_cpuxxx, &current_cpuyyy);
    set_procstat_value(current_cpuuser, old_cpuuser);
    set_procstat_value(current_cpunice, old_cpunice);
    set_procstat_value(current_cpusystem, old_cpusystem);
    set_procstat_value(current_cpuidle, old_cpuidle);
    set_procstat_value(current_cpuiowait, old_cpuiowait);
    set_procstat_value(current_cpuyyy, old_cpuyyy);
    set_procstat_value(current_cpuxxx, old_cpuxxx);
}
