/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "../../src/initng.h"

/*
   We should sort it as per filelist-order.cxx, see readahead-list
   - this should be handled by readahead_makecache?
   - keep in mind it then might not do it in the order we'd like.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <sched.h>
#include <sys/mman.h>
#include <errno.h>


#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_toolbox.h"

#include "../readahead_makecache/initng_readahead_database.h"

#include "initng_readahead.h"
#include "initng_iowait.h"


/* do the actual readahead */
static void process_file(char *file)
{
    int fd;
    struct stat st;

    /* TODO, need to have a minimum here, so we only readahead files who has more than X cache hits */
    /* We should also checkout doing up until 30 seconds after boot, then sleeping for 15 seconds, then continuing? */
    /* format of file is <seconds_after_boot> <cache hits> <filename> */
    if (strlen(file) > 23)
        file = &file[22];
    else
        return;
    D_("readahead %s\n", file);
    fd = open(file, O_RDONLY);
    if ((fd < 0) || (fstat(fd, &st) < 0))
        return;
    readahead(fd, (loff_t) 0, (size_t) st.st_size);
    close(fd);
    sched_yield();
}

static void readahead_database(const char *database)
{
    /* This code is based on readahead-list */
    int fd;
    struct stat st;
    char *file = NULL;
    int do_iowait_check = 0;
    char *iter;
    char *next;
    char buffer[MAXPATH + 1];

    fd = open(database, O_RDONLY);
    if ((fd < 0) || (fstat(fd, &st) < 0))
    {
        W_("Failed to open %s for readahead!\n", database);
        W_("this is ok if this is the first time you boot with readahead_makecache\n");
        W_("strerror : %s (%d)\n", strerror(errno), (int) fd);
        return;
    }
    file = mmap(0, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
    if (!file || file == MAP_FAILED)
    {
        F_("Failed mmap on %s\n", database);
        close(fd);
        return;
    }

    iter = file;

    while (iter)
    {
        /* find next newline */
        next = memchr(iter, '\n', file + st.st_size - iter);
        if (next)
        {
            if (next - iter < MAXPATH && (next - iter > 1))
            {
                memcpy(buffer, iter, next - iter);
                buffer[next - iter] = '\0';
                process_file(buffer);


                {
                    /* do an iowait check every 5 files (although wait atleast MIN_PROCTIME seconds in between checks, to let /proc/stat update) */
                    do_iowait_check++;
                    if (do_iowait_check > 5)
                    {
                        if (do_recheck())
                        {
                            collectstat();
                            W_("current_cpuiowait : %d\n", current_cpuiowait);
                            while (current_cpuiowait > 30)
                            {
                                usleep(MIN_PROCTIME);
                                collectstat();
                                W_("current_cpuiowait : %d\n",
                                   current_cpuiowait);
                            }
                            W_("Continuing\n");
                        }
                        do_iowait_check = 0;
                    }
                }
            }
            iter = next + 1;
        }
        else
        {
            iter = NULL;
        }

    }

    munmap(file, st.st_size);
    close(fd);
    W_("Readahead complete\n");
    module_unload();

}

int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    iowait_init();

#ifdef READAHEAD_BACKGROUND
    if (fork() == 0)
    {
#endif
        readahead_database(READAHEAD_CACHE_FILE);
#ifdef READAHEAD_BACKGROUND
        exit(0);
    }
#endif

    return TRUE;
}

void module_unload(void)
{
    D_("module_unload();\n");
    if (proc_stat_fd > 0)
    {
        close(proc_stat_fd);
        proc_stat_fd = 0;
    }
}
