
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
   on readahead ,scan through, find higest hit, and require number of hits to be 50% of that to do actual readahead on the file

   How this works (database part):
   Adds all entries to database during boot using the /proc filesystem.
   add_to_db
   Cleans database, this resolves symbolic links and after that clears duplicates
   clean_database
   Sorts database, sort is based on seconds_since_boot, to load things that will get accessed fast first.
   sort_database
   Loads and merges old results with current
   load_database (this uses merge_to_db) macro for merging data
   (
   Optionally trim's database, this removes all entries under MINIMUM_CACHE_HITS_ON_TRIM, and sets hits to hits / 2.
   trim_database
   )
   Saves new database to disk
   save_database
   Clears memory
   clear_database
 */

/*
   todo, make more recursive function out of for example readlink and duplicate check code? 
   cleanups, make a define for database_debug that shows more data?
   find out if we need a readahead.core file (from first boot, check every 1 second, recreate if a file in it is missing
   OR if readahead.dynamic, check say every 10 seconds is enough.
   check load averages on different checks, and boot times, see if checking every 5-10 seconds makes a difference in boot time.
   fix readahead plugin, and do the readahead on module_init, that can clean up the code a bit.
 */
#include "../../src/initng.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <sched.h>
#include <errno.h>
#include <sys/sysinfo.h>

#include "../../src/initng_toolbox.h"

#include "initng_readahead_makecache.h"
#include "initng_readahead_database.h"

/* This checks if an add_line is ok, dealing with filters and such */
static int add_line_check(char *name, int len)
{
    if (len < 2)
        return (FALSE);
    if (name[0] != '/')
        return (FALSE);
    if (name[len] == ')')
        return (FALSE);

    /* todo, make this more dynamic and user configurable */
    if (strncmp(name, "/proc", 5) == 0)
        return (FALSE);
    if (strncmp(name, "/dev", 4) == 0)
        return (FALSE);
    if (strncmp(name, "/tmp", 4) == 0)
        return (FALSE);
    if (strncmp(name, "/var/log", 8) == 0)
        return (FALSE);
    if (strncmp(name, "/var/tmp", 8) == 0)
        return (FALSE);
    if (strncmp(name, "/var/run", 8) == 0)
        return (FALSE);



    /* Make this optional? */
    /*    if (strncmp(name,"/mnt",4) == 0)
       return (FALSE); */
    if (strncmp(name, "/home", 5) == 0)
        return (FALSE);

    return (TRUE);
}

time_t get_time_since_boot(void)
{
    struct sysinfo si;

    sysinfo(&si);
    return si.uptime;
}

/* This adds line to database, unless it's there from before, if it is we increase hits and recalculate seconds_since_boot */
proc_data *add_to_db_real(char *name, int len, int seconds_since_boot,
                          int hits, proc_data * db)
{
    sched_yield();
    if (!add_line_check(name, len))
        return db;
    if (!db)
    {
        db = initng_calloc(1, sizeof(proc_data));
        //D_("Adding %s to database\n",name);
        db->name = strdup(name);
        db->hits = hits;
        db->seconds_since_boot = seconds_since_boot;
        return db;
    }
    if (db->name && (strcmp(name, db->name) == 0))
    {
        //D_("Cache hit for %s\n",name);
        db->hits += hits;
        db->seconds_since_boot =
            (db->seconds_since_boot <
             seconds_since_boot) ? db->
            seconds_since_boot : seconds_since_boot;
        return db;
    }
    db->next = add_to_db_real(name, len, seconds_since_boot, hits, db->next);
    return db;
}

/* This cleans up the database, doing a readlink so we dont have symbolic links, and removes duplicates */
void clean_database(void)
{
    char *readlink_buf = malloc(MAXPATH);       /* THIS IS NEEDED AS READLINK BUF, needs to be bigger? */
    int readlink_ret;
    struct stat st;
    proc_data *db = proc_db;
    proc_data *db2 = proc_db;

    if (!db)
    {
        D_("No proc db!\n");
        return;
    }

    /* Fix readlink to make sure we have the files, not the links to the files */
    while (db)
    {
        readlink_ret = readlink(db->name, readlink_buf, MAXPATH);
        while (readlink_ret > 0)
        {
            if (readlink_ret > MAXPATH)
            {
                F_("BUFFER NOT BIG ENOUGH FOR %s\n", db->name);
                continue;
            }
            if (readlink_ret > 0)
            {
                free(db->name);
                db->name = readlink_buf;
                readlink_buf = malloc(MAXPATH);
            }
            readlink_ret = readlink(db->name, readlink_buf, MAXPATH);
        }
        if (db->name && db->name[0] != '/')
        {
            free(db->name);
            db->name = NULL;
        }
        if (db->name)
        {
            if (lstat(db->name, &st) == 0)
            {
                /* Check if file is a file and not a directory */
                if (!S_ISREG(st.st_mode))
                {
                    free(db->name);
                    db->name = NULL;
                }
            }
            else
            {
                /* Failed to stat file */
                free(db->name);
                db->name = NULL;
            }
        }
        db = db->next;
    }
    free(readlink_buf);

    /* Check for duplicates */
    db = proc_db;
    while (db)
    {
        db2 = proc_db;
        while (db2)
        {
            if (db2 != db && db->name && db2->name
                && (strcmp(db->name, db2->name) == 0))
            {
                free(db->name);
                if (db->seconds_since_boot > db2->seconds_since_boot)
                    db2->seconds_since_boot = db->seconds_since_boot;
                db2->hits += db->hits;
                db->name = NULL;
            }
            db2 = db2->next;
        }
        db = db->next;
    }
}

/* This trims the database, removing files that have too few hits */
proc_data *trim_database_real(proc_data * db)
{
    proc_data *next;

    if (db)
    {
        if (db->next)
            db->next = trim_database_real(db->next);

        if (db->hits < MINIMUM_CACHE_HITS_ON_TRIM)
        {
            D_("Removing %s from database, too few cache hits (%d)\n",
               db->name, db->hits);
            if (db->name)
                free(db->name);
            next = db->next;
            free(db);
            return next;
        }
        else
        {
            /* We double seconds since boot, to make sure we remove any stuff that's potentionally started too early */
            db->seconds_since_boot = db->seconds_since_boot * 2;
            /* We half hits, to remove any old stuff */
            db->hits = db->hits / 2;
            if (db->hits > MINIMUM_CACHE_HITS_ON_TRIM)
                db->hits = MINIMUM_CACHE_HITS_ON_TRIM;
        }
    }
    return db;
}

/* This saves the current list to file */
void save_database(void)
{
    FILE *fp;
    int count = 0;
    proc_data *db = proc_db;

    mkdir(READAHEAD_CACHEDIR, S_IRUSR | S_IWUSR | S_IXUSR);
    if ((fp = fopen(READAHEAD_CACHE_FILE, "w")) == NULL)
    {
        F_("Failed to open %s for write\n", READAHEAD_CACHE_FILE);
        return;
    }
    fprintf(fp, "%10d\n", boots_since_last_trim);
    D_("Saving database:\n");
    D_("    %10d\n", boots_since_last_trim);
    while (db)
    {
        if (db->name)
        {
            fprintf(fp, "%10d %10d %s\n", db->seconds_since_boot, db->hits,
                    db->name);
            D_("    %10d %10d %s\n", db->seconds_since_boot, db->hits,
               db->name);
            count++;
        }
        db = db->next;
    }

    fclose(fp);
    D_("Number of entries saved : %d\n", count);
}

void load_database(void)
{
    FILE *fp;
    int count = 0;
    char buf[MAXPATH + 30];
    char *tmp;
    int hits;
    int seconds_since_boot;

    if ((fp = fopen(READAHEAD_CACHE_FILE, "r")) == NULL)
    {
        F_("Failed to open %s for read\n", READAHEAD_CACHE_FILE);
        F_("This is ok if this is the first time your booting with readahead_makecache\n");
        F_("strerror : %s\n", strerror(errno));
        return;
    }
    if (!fgets(buf, sizeof(buf), fp))
    {
        F_("Failed to load database!\n");
        return;
    }
    boots_since_last_trim = atoi(buf);
    buf[10] = 0;
    D_("Boots since last trim from database : %d\n", boots_since_last_trim);
    while (fgets(buf, sizeof(buf), fp))
    {
        /* remove \n */
        buf[strlen(buf) - 1] = 0;
        buf[10] = 0;
        buf[21] = 0;
        tmp = &buf[0];
        seconds_since_boot = atoi(tmp);
        tmp = &buf[11];
        hits = atoi(tmp);
        tmp = &buf[22];
        D_("Adding %s with %d hits and %d seconds since boot\n", tmp, hits,
           seconds_since_boot);
        merge_to_db(tmp, seconds_since_boot, hits);
        count++;
    }
    fclose(fp);
    D_("Number of entries loaded : %d\n", count);
}

static proc_data *sort_database_entry(proc_data * db, int *db_changed)
{
    proc_data *tmp;

    if (db->next)
    {
        if (db->next->seconds_since_boot < db->seconds_since_boot)
        {
            //D_("%s should be earlier than %s\n",db->next->name,db->name);
            /* Switch places on them */
            tmp = db->next;
            db->next = db->next->next;
            tmp->next = db;
            db = tmp;

            *db_changed = TRUE;
            /*            {   // debug block 
               printf("Returning %s\n",db->name);
               tmp = db->next;
               while (tmp) {
               printf("  next : %s\n",tmp->name);
               tmp = tmp->next;
               }
               } */
            return db;
        }
        db->next = sort_database_entry(db->next, db_changed);
    }
    return db;

}

void sort_database(void)
{
    int db_changed = TRUE;

    D_("Scanning db\n");
    while (db_changed)
    {
        db_changed = FALSE;
        proc_db = sort_database_entry(proc_db, &db_changed);
    }
}

void free_database_real(proc_data * db)
{
    if (!db)
        return;
    if (db->next)
        free_database_real(db->next);
    if (db->name)
        free(db->name);
    free(db);
}
